import typing
from datetime import date, timedelta

from .errors import (
    AcctNumberMissingError, AcctNumberInvalidError, AcctNumberTypeError,
    ABIMissingError, ABIInvalidError, ABITypeError,
    CABMissingError, CABInvalidError, CABTypeError,
    DuedateMissingError, DuedateTooEarlyError, DuedateTypeError,
    SIAMissingError, SIATypeError, SIAInvalidError,
    ZIPInvalidError, ZIPTypeError
)


# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

def validate_abi(abi: typing.Union[str, int, bool, None], nome_soggetto: str):

    if abi == '' or abi is False or abi is None:
        raise ABIMissingError(f'Codice ABI mancante per "{nome_soggetto}"')

    elif isinstance(abi, int):
        validate_abi(str(abi).rjust(5, '0'), nome_soggetto)

    elif isinstance(abi, str):
        if not abi.isnumeric():
            raise ABIInvalidError(
                f'Codice ABI errato per "{nome_soggetto}".'
                f' - '
                f'Il codice ABI deve essere un numero di 5 cifre'
            )

        elif not len(abi) == 5:
            raise ABIInvalidError(
                f'Codice ABI errato per "{nome_soggetto}".'
                f' - '
                f'Il codice ABI deve essere un numero di 5 cifre'
            )

        elif int(abi) == 0:
            raise ABIInvalidError(
                f'Codice ABI errato per "{nome_soggetto}".'
                f' - '
                f'Il codice ABI non può essere composto da 5 zeri (00000)')

        else:
            # Codice ABI OK
            return
        # end if

    else:
        raise ABITypeError(
            f'Tipo dato non valido ({type(abi)}) '
            f'per il codice ABI di "{nome_soggetto}".'
            f' - '
            f'Il tipo può essere stringa (str) o numero intero (int)'
        )
    # end if
# end validate_abi


def validate_cab(cab: typing.Union[str, int, bool, None], nome_soggetto: str):

    if cab == '' or cab is False or cab is None:
        raise CABMissingError(f'Codice CAB mancante per "{nome_soggetto}"')
    
    elif isinstance(cab, int):
        validate_cab(str(cab).rjust(5, '0'), nome_soggetto)

    elif cab in ('', False, None):
        raise CABMissingError(f'Codice CAB mancante per "{nome_soggetto}"')

    elif isinstance(cab, str):
        if not cab.isnumeric():
            raise CABInvalidError(
                f'Codice CAB errato per "{nome_soggetto}".'
                f' - '
                f'Il codice CAB deve essere un numero di 5 cifre'
            )

        elif not len(cab) == 5:
            raise CABInvalidError(
                f'Codice CAB errato per "{nome_soggetto}".'
                f' - '
                f'Il codice CAB deve essere un numero di 5 cifre'
            )

        elif int(cab) == 0:
            raise CABInvalidError(
                f'Codice CAB errato per "{nome_soggetto}".'
                f' - '
                f'Il codice CAB non può essere composto da 5 zeri (00000)')

        else:
            # Codice CAB OK
            return
        # end if

    else:
        raise CABTypeError(
            f'Tipo dato non valido ({type(cab)}) '
            f'per il codice CAB di "{nome_soggetto}".'
            f' - '
            f'Il tipo può essere stringa (str) o numero intero (int)'
        )
    # end if
# end validate_cab


def validate_bank_account_number(
    account_number: typing.Union[str, int, bool, None]
):
    acct_num = account_number
    if acct_num == '' or acct_num is False or acct_num is None:
        raise AcctNumberMissingError(
            'Numero di conto corrente del creditore mancante'
        )
    
    elif isinstance(account_number, int):
        acc_number_str = str(account_number).rjust(12, '0')
        validate_bank_account_number(acc_number_str)

    elif isinstance(account_number, str):
        acc_number_str = str(account_number)
        
        # Pick the last 12 characters to correctly manage italian IBAN passed
        # as account numbers
        if len(acc_number_str) == 27 and acc_number_str.startswith('IT'):
            acc_number_str = acc_number_str[-12:]
        else:
            acc_number_str = account_number
        # end if
        
        if not acc_number_str.isnumeric():
            raise AcctNumberInvalidError(
                f'Numero di contro corrente del creditore errato'
                f' - '
                f'Il contro coorente deve essere un numero di 12 cifre'
            )

        elif not len(acc_number_str) == 12:
            raise AcctNumberInvalidError(
                f'Numero di contro corrente del creditore errato'
                f' - '
                f'Il contro coorente deve essere un numero di 12 cifre'
            )

        elif int(acc_number_str) == 0:
            raise AcctNumberInvalidError(
                f'Numero di contro corrente del creditore errato'
                f' - '
                f'Il numero del conto non può essere composto solo da zeri')

        else:
            # Codice CAB OK
            return
        # end if

    else:
        raise AcctNumberTypeError(
            f'Tipo dato non valido ({type(account_number)}) '
            f'per il numero di conto corrente del creditore.'
            f' - '
            f'Il tipo può essere stringa (str) o numero intero (int)'
        )
    # end if
# end validate_bank_account_number


def validate_duedate(duedate: typing.Union[date, bool, None], nome_soggetto: str):
    
    tomorrow = date.today() + timedelta(days=1)
    
    if not duedate:
        raise DuedateMissingError(
            f'Data di scadenza mancante per "{nome_soggetto}"'
        )
    
    elif not isinstance(duedate, date):
        raise DuedateTypeError(
            f'Tipo dato non valido ({type(duedate)}) '
            f'per la data di scadenza di "{nome_soggetto}".'
            f' - '
            f'Il tipo deve essere datetime.date'
        )
    
    elif not duedate >= tomorrow:
        raise DuedateTooEarlyError(
            f'Data scadenza pagamento ({duedate}) troppo vicina nel tempo '
            f'per "{nome_soggetto}".'
            f' - '
            f'La data discadenza non può essere prima di domani.'
        )
    
    else:
        pass
    # end if
# end validate_duedate


def validate_sia(sia_code: typing.Union[str, bool, None]) -> None:
    # Verifica SIA: campo obbligatorio. Il formato sia prevede una
    # lettera seguita da 4 numeri

    if sia_code == '' or sia_code is False or sia_code is None:
        raise SIAMissingError(
            'Codice SIA del creditore mancante'
        )

    elif not isinstance(sia_code, str):
        raise SIATypeError(
            'Tipo dato non valido per il codice SIA: '
            'il tipo deve essere stringa (str)'
        )

    elif not len(sia_code) == 5:
        raise SIAInvalidError(
            'Codice SIA errato: il codice deve essere di lunghezza 5')

    elif not sia_code[0].isalpha():
        raise SIAInvalidError(
            'Codice SIA errato: il primo componente deve essere una lettera')

    elif not sia_code[1:].isnumeric():
        raise SIAInvalidError(
            'Codice SIA errato: gli ultimi 4 componenti devono essee numeri')

    else:
        # Codice SIA OK
        pass
    # end if
# end validate zip


def validate_zip(zip_code: typing.Union[str, int, bool, None]) -> None:
    # Verifica CAP: se c'è deve essere numerico e di 5 caratteri,
    # ma può tranquillamente non esserci!!

    if zip_code == '' or zip_code is False or zip_code is None:
        # Empty ZIP code is OK!
        return

    elif isinstance(zip_code, int):
        validate_zip(str(zip_code).rjust(5, '0'))

    elif isinstance(zip_code, str):
        if not zip_code.isnumeric():
            raise ZIPInvalidError(
                'CAP errato. Il CAP deve essere un numero di 5 cifre')

        elif len(zip_code) != 5:
            raise ZIPInvalidError(
                'CAP errato. Il CAP deve essere un numero di 5 cifre')

        elif int(zip_code) == 0:
            raise ZIPInvalidError(
                'CAP errato. Il CAP non può essere composto da 5 zeri (00000)')

        else:
            # Codice ZIP OK
            return
        # end if

    else:
        raise ZIPTypeError(
            f'Tipo dato non valido per il CAP: {type(zip_code)}'
            f' - '
            f'Il tipo può essere stringa (str) o numero intero (int)'
        )
    # end if
# end validate zip



from .odoo_stuff import UserError


class AcctNumberMissingError(UserError):
    pass
# end AcctNumberMissingError


class AcctNumberInvalidError(UserError):
    pass
# end AcctNumberInvalidError


class AcctNumberTypeError(UserError):
    pass
# end AcctNumberTypeError


class ABIMissingError(UserError):
    pass
# end ABIMissingError


class ABIInvalidError(UserError):
    pass
# end ABIInvalidError


class ABITypeError(UserError):
    pass
# end ABITypeError


class PartnerBankMissing(UserError):
    pass
# end PartnerBankMissing


class CABMissingError(UserError):
    pass
# end CABMissingError


class CABInvalidError(UserError):
    pass
# end CABInvalidError


class CABTypeError(UserError):
    pass
# end CABTypeError


class DuedateMissingError(UserError):
    pass
# end DuedateMissingError


class DuedateTooEarlyError(UserError):
    pass
# end DuedateTooEarlyError


class DuedateTypeError(UserError):
    pass
# end DuedateMissingError


class FiscalcodeMissingError(UserError):
    pass
# end CABTypeError


class SIAMissingError(UserError):
    pass
# end SIAMissingError


class SIATypeError(UserError):
    pass
# end SIATypeError


class SIAInvalidError(UserError):
    pass
# end SIAInvalidError


class ZIPInvalidError(UserError):
    pass
# end ZIPInvalidError


class ZIPTypeError(UserError):
    pass
# end ZIPTypeError



import itertools
from datetime import datetime, date
import importlib.resources

import typing
from mako.template import Template

from .utils.errors import FiscalcodeMissingError, PartnerBankMissing
from .utils.odoo_stuff import _
from .utils.validators import (
    validate_abi,
    validate_cab,
    validate_bank_account_number,
    validate_zip,
    validate_sia
)


# Name of the Mako template file
CBI_TEMPLATE_FILE = 'cbi.mako'


# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Receipt:
    """
    Class that represents a single RiBa receipt to be added to a RiBa document.
    
    :param duedate_move_line: Odoo object representing the amount to be payed and it's maturity date
    :type duedate_move_line: class:`account.move.line`
    :param invoice: Odoo object representing the invoice of which this receipt is part
    :type invoice: class:`account.invoice`
    :param debtor_partner: Odoo object holding the name and address of the debtor
    :type debtor_partner: class:`res.partner`
    :param debtor_bank: Odoo object holding the name and address of the debtor's bank
    :type debtor_bank: class:`res.bank`
    """

    def __init__(self, duedate_move_line, invoice, debtor_partner, debtor_bank):

        # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
        # Fields initialization
        # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

        self._duedate_move_line = duedate_move_line
        self._invoice = invoice
        self._debtor_partner = debtor_partner
        self._debtor_bank = debtor_bank

        # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
        # Sanity checks
        # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

        # CAB and ABI required
        if not self._debtor_bank:
            raise PartnerBankMissing(
                f'No bank specified for {self.debtor_name}'
            )
        # end if
        abi = self.debtor_bank.abi
        cab = self.debtor_bank.cab

        validate_abi(abi, self.debtor_name)
        validate_cab(cab, self.debtor_name)

        # Fiscal code required
        if not self.debtor_fiscalcode:
            raise FiscalcodeMissingError(
                _('Fiscal Code not specified for ') + self.debtor_name
            )
        # end if

        # Validate ZIP code
        validate_zip(self.debtor_zip)

    # end __init__
    
    @property
    def is_group(self):
        return False
    # end is_group

    @property
    def duedate(self):
        return self._duedate_move_line.date_maturity
    # end duedate

    @property
    def amount(self):
        return self._duedate_move_line.amount_residual
    # end amount

    @property
    def debtor_partner(self):
        return self._debtor_partner
    # end duedate

    @property
    def debtor_name(self):
        return self._debtor_partner.name
    # end debtor_name

    @property
    def debtor_client_code(self):
        return self._debtor_partner.ref or ''
    # end debtor_client_code

    @property
    def debtor_fiscalcode(self):
        return self._debtor_partner.fiscalcode
    # end debtor_fiscalcode

    @property
    def debtor_vat_number(self):
        if not self._debtor_partner.vat:
            return False
        else:
            return self._debtor_partner.vat
        # end if
    # end debtor_vat_number

    @property
    def debtor_fiscode_or_vat(self):
        return self.debtor_fiscalcode or self.debtor_vat_number
    # end debtor_fiscode_or_vat

    @property
    def debtor_address(self):
        return self._debtor_partner.street
    # end debtor_address

    @property
    def debtor_city(self):
        return self._debtor_partner.city
    # end debtor_city

    @property
    def debtor_state(self):
        if self._debtor_partner.state_id:
            return str(self._debtor_partner.state_id.code)
        else:
            return ''
        # end if
    # end debtor_state

    @property
    def debtor_zip(self):
        return self._debtor_partner.zip
    # end debtor_zip

    @property
    def debtor_bank(self):
        return self._debtor_bank
    # end debtor_bank

    @property
    def invoice_number(self):
        if self._invoice:
            return self._invoice.number
        else:
            return self._duedate_move_line.move_id.name
    # end invoice_number

    @property
    def invoice_date(self):
        if self._invoice:
            return self._invoice.date_invoice
        else:
            return self._duedate_move_line.move_id.invoice_date
    # end invoice_date
    
    @property
    def grouping_key(self) -> typing.Tuple[str, str, str, date]:
        """
        Generate and return the grouping key used to group the class:`Receipt` objects when requested
        
        :returns: The grouping key
        :rtype: Tuple
        """
        
        return (
            str(self.debtor_partner.id),
            str(self._debtor_bank.abi),
            str(self._debtor_bank.cab),
            self.duedate
        )
    # end gkey
    
    def __str__(self):
        name = self.debtor_name[:25].ljust(25, ' ')
        vat = self.debtor_fiscode_or_vat
        inv_num = self.invoice_number
        inv_date = self.invoice_date
        due_date = self.duedate
        amount = self.amount
        return f'{name} ({vat}) - ' \
               f'{inv_num} del {inv_date:%Y-%m-%d} - ' \
               f'scad: {due_date:%Y-%m-%d} € {amount}'
    # end __str__
    
    def __repr__(self):
        return self.__str__()
    # end __repr__
# end Receipt


class ReceiptGroup:
    """
    Class that represents a group of RiBa receipt to be rendered as a single line in the RiBa document.
    
    :param receipts: the list of class:`Receipt` objects to bo grouped together
    :type receipts: List of class:`Receipt`
    """

    def __init__(self, receipts: typing.List[Receipt]):
        
        if not receipts:
            raise ValueError(
                'There must be at least one Receipt object in the group'
            )
        # end if

        # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
        # Fields initialization
        # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
        # List of receipts
        self._receipts = receipts
        
        # Reference object to get data common to all receipts in the group
        self._r_ref = self._receipts[0]
        
        # Description of the group listing the numbers of the invoices
        # referred by the grouped lines
        self._desc = ', '.join(
            map(lambda r: f'{r.invoice_number} ({r.invoice_date:%Y-%m-%d}) € {r.amount:.2f}', self._receipts)
        )
        
        # Total amount of the group
        self._amount = sum(
            map(lambda r: r.amount, self._receipts)
        )

        # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
        # Sanity checks
        # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
        
        # Same debtor partner
        
        # Same debtor bank
    # end __init__
    
    @property
    def is_group(self):
        return len(self._receipts) > 1
    # end is_group
    
    @property
    def grouped_receipts(self):
        return self._receipts
    # end is_group

    @property
    def duedate(self):
        return self._r_ref.duedate
    # end duedate

    @property
    def amount(self):
        return self._amount
    # end amount

    @property
    def debtor_name(self):
        return self._r_ref.debtor_name
    # end debtor_name

    @property
    def debtor_client_code(self):
        return self._r_ref.debtor_client_code
    # end debtor_client_code

    @property
    def debtor_fiscalcode(self):
        return self._r_ref.debtor_fiscalcode
    # end debtor_fiscalcode

    @property
    def debtor_vat_number(self):
        return self._r_ref.debtor_vat_number
    # end debtor_vat_number

    @property
    def debtor_fiscode_or_vat(self):
        return self._r_ref.debtor_fiscode_or_vat
    # end debtor_fiscode_or_vat

    @property
    def debtor_address(self):
        return self._r_ref.debtor_address
    # end debtor_address

    @property
    def debtor_city(self):
        return self._r_ref.debtor_city
    # end debtor_city

    @property
    def debtor_state(self):
        return self._r_ref.debtor_state
    # end debtor_state

    @property
    def debtor_zip(self):
        return self._r_ref.debtor_zip
    # end debtor_zip

    @property
    def debtor_bank(self):
        return self._r_ref.debtor_bank
    # end debtor_bank

    @property
    def invoice_number(self):
        return self._r_ref.invoice_number
    # end invoice_number

    @property
    def invoice_date(self):
        return self._r_ref.invoice_date
    # end invoice_date
    
    def __str__(self):
        d_name = self.debtor_name[:25].ljust(25, ' ')
        return f'{d_name} ({self.debtor_fiscode_or_vat}) - ' \
               f'{self.duedate:%Y-%m-%d} - {self._desc} - {self.amount} €'
    # end __str__
    
    def __repr__(self):
        return self.__str__()
    # end __repr__
# end ReceiptGroup


class Document:
    """
    Class that represents a RiBa document with header record, trailing record
    and records for RiBa receipts.
    
    :param creditor_company: Odoo object representing the creditor's company
    :type creditor_company: class:`res.company`
    :param creditor_bank_account:
    :type creditor_company: class:`res.partner.bank`
    """

    def __init__(self, creditor_company, creditor_bank_account):
        """Constructor method"""

        # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
        # Fields initialization
        # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

        self._creditor_company = creditor_company
        self._creditor_bank_account = creditor_bank_account
        self._creation_date = datetime.now()

        self._receipts = list()

        # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
        # Sanity checks
        # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

        # CAB and ABI required
        abi = self.creditor_bank_account.bank_abi
        cab = self.creditor_bank_account.bank_cab

        validate_abi(abi, self.creditor_company_name)
        validate_cab(cab, self.creditor_company_name)

        validate_bank_account_number(
            self.creditor_bank_account.sanitized_acc_number
        )

        validate_sia(self.sia_code)

        if not self.creditor_fiscode_or_vat:
            raise FiscalcodeMissingError(
                _('No VAT or Fiscal Code specified for ') + self.creditor_company_name
            )
        # end if

        # Validate ZIP code
        validate_zip(self.creditor_company_addr_zip)
    # end __init__

    @property
    def creditor_company(self):
        return self._creditor_company
    # end creditor_company

    @property
    def creditor_company_name(self):
        return self._creditor_company.partner_id.name
    # end creditor_company

    @property
    def creditor_fiscalcode(self):
        return self._creditor_company.partner_id.fiscalcode
    # end creditor_fiscalcode

    @property
    def creditor_vat_number(self):
        if not self._creditor_company.partner_id.vat:
            return False
        else:
            return self._creditor_company.partner_id.vat
        # end if
    # end creditor_vat_number

    @property
    def creditor_fiscode_or_vat(self):
        return self.creditor_fiscalcode or self.creditor_vat_number
    # end creditor_fiscode_or_vat

    @property
    def creditor_company_ref(self):
        return self._creditor_company.partner_id.ref or ''
    # end creditor_company

    @property
    def creditor_company_addr_street(self):
        return self._creditor_company.partner_id.street or ''
    # end creditor_company

    @property
    def creditor_company_addr_zip(self):
        zip_code = self._creditor_company.partner_id.zip or ''
        return zip_code
    # end creditor_company

    @property
    def creditor_company_addr_city(self):
        city = self._creditor_company.partner_id.city or ''
        return city
    # end creditor_company

    @property
    def creditor_company_addr_zip_and_city(self):
        zip_code = self._creditor_company.partner_id.zip or ''
        city = self._creditor_company.partner_id.city or ''
        return f'{zip_code} {city}'
    # end creditor_company

    @property
    def creditor_bank_account(self):
        return self._creditor_bank_account
    # end creditor_bank_account

    @property
    def creation_date(self):
        return self._creation_date
    # end creation_date

    @property
    def sia_code(self):
        return str(self._creditor_company.sia_code).strip()
    # end sia_code

    @property
    def name(self):
        """Return the "Nome supporto" as specified in the reference
         document CBI-ICI-001 Versione: v. 6.01"""
        return self._creation_date.strftime('%d%m%y%H%M%S') + str(self.sia_code)
    # end support_name

    @property
    def total_amount(self):
        # Since this function will probably be called just one time the
        # computation can be done on the fly without any negative
        # performance impact
        receipts_amounts = map(lambda line: line.amount, self._receipts)
        total_amount = sum(receipts_amounts)
        return total_amount
    # end total_amount

    def add_receipt(self, rcpt: Receipt):
        """
        Add a receipt to the RiBa document
        :return: nothing
        """
        self._receipts.append(rcpt)
    # end add_line

    def render_cbi(self, group: bool = False):
        """
        Render the RiBa document in the CBI format
        :return: the CBI document representing the RiBa document
        :rtype: str
        """
        
        # Load the Mako template
        cbi_template = Template(
            text=importlib.resources.read_text(__package__ + '.templates', CBI_TEMPLATE_FILE)
        )
        
        # Group the receipts if requested
        if group:
            receipt_groups = self._group_receipts()
            cbi_document = cbi_template.render(doc=self, lines=receipt_groups)
        else:
            cbi_document = cbi_template.render(doc=self, lines=self._receipts)
        # end if
        return cbi_document
    # end render
    
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    # Private methods
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    def _group_receipts(self):
        
        # Sort the receipts
        def key_f(x): return x.grouping_key
        
        # Sort the receipts
        rcpt_sorted = sorted(self._receipts, key=key_f)
        
        # Group the receipts (same debtor and same duedate)
        rcpt_groups = [
            # 'g' must be converted to a list because it's just an iterator
            ReceiptGroup(list(g))
            for _, g in itertools.groupby(rcpt_sorted, key=key_f)
        ]
        
        return rcpt_groups
    # end _group_receipts
# end Document



# Try importing odoo functions and classes, if import fails add define for
# the required functions and classes
try:
    from odoo.exceptions import UserError
except (ModuleNotFoundError, ImportError) as err:
    class UserError(Exception):
        pass
    # end UserError
# end try / except

try:
    from odoo import _
except (ModuleNotFoundError, ImportError) as err:
    def _(val):
        return val
    # end if
# end try / except



#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['ribalta', 'ribalta.templates', 'ribalta.utils']

package_data = \
{'': ['*']}

package_dir = \
{'': 'src'}

install_requires = \
['mako >=1.1.3', 'unidecode']

extras_require = \
{'test': ['pytest >=2.7.3', 'pytest-cov', 'python-dateutil']}

setup(name='ribalta',
      version='0.2.6',
      description='RiBa tools!',
      author='Marco Tosato',
      author_email='marco.tosato@didotech.com',
      url='https://github.com/LibrERP/Pylibs/',
      packages=packages,
      package_data=package_data,
      package_dir=package_dir,
      install_requires=install_requires,
      extras_require=extras_require,
      python_requires='>= 3.6',
     )
